//#define BENCHMARK_OFF //  .
//#define BENCHMARK_WITHOUT_OVERHEAD //       .

#ifndef BENCHMARK_GETMICROSECONDS_MAXINTERVAL
  #define BENCHMARK_GETMICROSECONDS_MAXINTERVAL 5
#endif // #ifndef BENCHMARK_GETMICROSECONDS_MAXINTERVAL

#define BENCHMARK_RESERVE 50

#ifdef __MQL5__

class BENCHMARK
{
private:

#endif // #ifdef __MQL5__

  struct BENCHMARK_STRUCT
  {
    ulong StartTime;
    string Str;

    int OverHead;
    int OverHeadAmount;
    int OverHeadGetMscCount;

    void Set( const string &sStr )
    {
      this.Str = sStr;

      this.OverHead = 0;
      this.OverHeadAmount = 0;
      this.OverHeadGetMscCount = 0;

      this.StartTime = ::GetMicrosecondCount();

      return;
    }

    ulong Get( void ) const
    {
      return(::GetMicrosecondCount() - this.StartTime);
    }

    string ToString( const ulong &Res ) const
    {


      return("Time[" + this.Str + "] = " +
                                         #ifdef BENCHMARK_WITHOUT_OVERHEAD
                                           (this.OverHeadGetMscCount ?
                                            (string)Res + " - " +
                                            (string)this.OverHead +
                                            " (" + (string)this.OverHeadAmount + " Msg, " + (string)this.OverHeadGetMscCount + " GMsc) = "
                                                                     : NULL) +
                                            (string)(Res - this.OverHead) + " mcs.");
                                         #else // #ifdef BENCHMARK_WITHOUT_OVERHEAD
                                           (this.OverHeadGetMscCount ?
                                             (string)(Res - this.OverHead) + " + " +
                                             (string)this.OverHead +
                                             " (" + (string)this.OverHeadAmount + " Msg, " + (string)this.OverHeadGetMscCount + " GMsc) = "
                                                                     : NULL) +
                                           (string)Res + " mcs.");
                                         #endif // #ifdef BENCHMARK_WITHOUT_OVERHEAD #else
    }

    void operator +=( const ulong Value )
    {
      this.OverHead += (int)(Value - this.StartTime);
      this.OverHeadAmount++;
      this.OverHeadGetMscCount++;

      return;
    }

    void operator +=( const BENCHMARK_STRUCT &Bnch )
    {
      this.OverHead += Bnch.OverHead;
      this.OverHeadAmount += Bnch.OverHeadAmount;
      this.OverHeadGetMscCount += Bnch.OverHeadGetMscCount + 2;

      return;
    }
  };

#ifndef __MQL5__

class BENCHMARK
{
private:

#endif // #ifndef __MQL5__

  static BENCHMARK_STRUCT Bench[];

  static int ReserveSize;
  static int Amount;

  static ulong GetAvgInterval( const int Attempts = 10 )
  {
    const ulong StartTime = ::GetMicrosecondCount();
    ulong NewTime = StartTime;

    for (int i = 0; i < Attempts; i++)
      NewTime = ::GetMicrosecondCount();

    return((NewTime - StartTime) / Attempts);
  }

  //  true,   VPS.
  static bool IsFreezeVPS( const int Attempts = 10 )
  {
  #ifdef GetMicrosecondCount
    ::Alert(__FILE__ + ": #" + "ifdef GetMicrosecondCount - true!");
  #endif // #ifdef GetMicrosecondCount

    //   -   .
    const ulong AvgInterval = BENCHMARK::GetAvgInterval(Attempts);
    const bool Res = (AvgInterval > (BENCHMARK_GETMICROSECONDS_MAXINTERVAL));

    if (Res)
      ::Alert(__FILE__ + ": FreezeVPS - TimeAvg[GetMicrosecondsCount()] = " + (string)AvgInterval + " mcs.");
    else
      ::Print(__FILE__ + ": TimeAvg[GetMicrosecondsCount()] = " + (string)AvgInterval + " mcs.");

  #ifdef BENCHMARK_OFF
    ::Print(__FILE__ + " is Off.");
  #else // #ifdef BENCHMARK_OFF
    ::Print(__FILE__ + " is On.");
  #endif // #ifdef BENCHMARK_OFF #else

    return(Res);
  }

public:
  static const bool IsNotFreezeVPS;

  static bool Set( const string Str )
  {
    if (BENCHMARK::Amount == BENCHMARK::ReserveSize)
      BENCHMARK::ReserveSize = ::ArrayResize(BENCHMARK::Bench, BENCHMARK::ReserveSize + BENCHMARK_RESERVE);

    BENCHMARK::Bench[BENCHMARK::Amount++].Set(Str);

    return(true);
  }

  static ulong Get( const uint AlertInterval = 0 )
  {
    const int Pos = BENCHMARK::Amount - 1;
    const ulong Res = (Pos < 0) ? 0 : BENCHMARK::Bench[Pos].Get();

    if (Pos >= 0)
    {
      if (AlertInterval && (Res > AlertInterval))
      {
      #ifndef Alert
        if (Pos && ((uint)BENCHMARK::Amount * 1000 > Res)) // https://www.mql5.com/ru/forum/342090/page47#comment_18631545
          ::Print("Bench_Stack = " + (string)Pos + ", " + (string)AlertInterval + " <= " + BENCHMARK::Bench[Pos].ToString(Res));
        else
      #endif // #ifndef Alert
          ::Alert("Bench_Stack = " + (string)Pos + ", " + (string)AlertInterval + " <= " + BENCHMARK::Bench[Pos].ToString(Res));

        if (Pos)
          BENCHMARK::Bench[Pos] += ::GetMicrosecondCount() - Res;
      }

      if (Pos)
        BENCHMARK::Bench[Pos - 1] += BENCHMARK::Bench[Pos];

      BENCHMARK::Amount = Pos;
    }

    return(Res);
  }

  template <typename T>
  static T Benchmark( const ulong Get, const T FuncValue, const bool Set )
  {
    return(FuncValue);
  }

  template <typename T>
  static T Benchmark( const ulong Get, const T &FuncValue, const bool Set )
  {
    return(FuncValue);
  }
};

#ifdef __MQL5__
  static BENCHMARK::BENCHMARK_STRUCT BENCHMARK::Bench[];
#else // __MQL5__
  static BENCHMARK_STRUCT BENCHMARK::Bench[];
#endif // __MQL5__

// static const bool BENCHMARK::IsNotFreezeVPS = true; //   ,    .
 static const bool BENCHMARK::IsNotFreezeVPS = !BENCHMARK::IsFreezeVPS();

static int BENCHMARK::ReserveSize = BENCHMARK::IsNotFreezeVPS ? 0 : ::ArrayResize(BENCHMARK::Bench, BENCHMARK_RESERVE);
static int BENCHMARK::Amount = 0;

#ifdef BENCHMARK_OFF
  // void
  #define _BV(A,B) A;

  // Not void
  #define _B(A,B) (A)
#else // #ifdef BENCHMARK_OFF
  // void
  #define _BV(A,B)                                                                          \
    if ((B) && BENCHMARK::IsNotFreezeVPS)                                                   \
    {                                                                                       \
      BENCHMARK::Set(__FILE__ + " " + (string)__LINE__ + " in " + __FUNCTION__+ ": " + #A); \
      A;                                                                                    \
      BENCHMARK::Get(B);                                                                    \
    }                                                                                       \
    else                                                                                    \
    {                                                                                       \
      A;                                                                                    \
    }

  // Not void
  #define _B(A,B) (((B) && BENCHMARK::IsNotFreezeVPS) ?                                                                                                  \
                   BENCHMARK::Benchmark(BENCHMARK::Get(B), A, BENCHMARK::Set(__FILE__ + " " + (string)__LINE__ + " in " + __FUNCTION__+ ": " + #A)) : A)

#endif // #ifdef BENCHMARK_OFF #else

#undef BENCHMARK_RESERVE